from pathlib import Path


def test_multiple_file_extensions(testdir, plugin_args_fails_xdist):
    file_extension = "ext2.ext1"

    testcase = f"""
    import pytest
    from syrupy.extensions.single_file import SingleFileSnapshotExtension

    class DotInFileExtension(SingleFileSnapshotExtension):
        _file_extension = "{file_extension}"

    @pytest.fixture
    def snapshot(snapshot):
        return snapshot.use_extension(DotInFileExtension)

    def test_dot_in_filename(snapshot):
        assert b"expected_data" == snapshot
    """

    test_file: Path = testdir.makepyfile(test_file=testcase)

    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0

    snapshot_file = (
        Path(test_file).parent
        / "__snapshots__"
        / "test_file"
        / f"test_dot_in_filename.{file_extension}"
    )
    assert snapshot_file.exists()

    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot passed\.",))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0


def test_class_style(testdir, plugin_args_fails_xdist):
    """
    Regression test for https://github.com/syrupy-project/syrupy/issues/717
    """

    testcase = """
    import pytest
    from syrupy.extensions.json import JSONSnapshotExtension

    @pytest.fixture
    def snapshot(snapshot):
        return snapshot.use_extension(JSONSnapshotExtension)

    class TestFoo:
        def test_foo(self, snapshot):
            assert { 'key': 'value' } == snapshot
    """

    test_file: Path = testdir.makepyfile(test_file=testcase)

    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot generated\.",))
    assert "deleted" not in result.stdout.str()
    assert result.ret == 0

    snapshot_file = (
        Path(test_file).parent / "__snapshots__" / "test_file" / "TestFoo.test_foo.json"
    )
    assert snapshot_file.exists()

    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot passed\.",))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0



from types import MappingProxyType

import pytest


def test_snapshot_diff(snapshot):
    my_dict = {
        "field_0": True,
        "field_1": "no_value",
        "nested": {
            "field_0": 1,
        },
    }
    assert my_dict == snapshot
    my_dict["field_1"] = "yes_value"
    assert my_dict == snapshot(diff=0)
    my_dict["nested"]["field_0"] = 2
    assert my_dict == snapshot(diff=0)


def test_snapshot_diff_id(snapshot):
    my_dict = {
        "field_0": True,
        "field_1": "no_value",
        "field_2": 0,
        "field_3": None,
        "field_4": 1,
        "field_5": False,
        "field_6": (True, "hey", 2, None),
        "field_7": {False, "no", None},
    }
    dict_large_snapshot = {
        **my_dict,
        "nested_0": dict(my_dict),
        "nested_1": dict(my_dict),
    }

    assert dict_large_snapshot == snapshot(name="large snapshot")
    dict_diff_large_snapshot = {
        **dict_large_snapshot,
        "field_1": "yes_value",
        "field_6": ("hey", 2, False),
        "field_7": {"yes", 0, None},
    }
    assert dict_diff_large_snapshot == snapshot(diff="large snapshot")
    dict_case_3 = {
        **dict_large_snapshot,
        "nested_0": MappingProxyType({**my_dict, "field_0": 2}),
        "nested_1": MappingProxyType({**my_dict, "field_0": 2}),
    }
    assert dict_case_3 == snapshot(name="case3", diff="large snapshot")


@pytest.mark.xfail(reason="Asserting snapshot does not exist")
def test_snapshot_no_diff_raises_exception(snapshot):
    my_dict = {
        "field_0": "value_0",
    }
    assert my_dict == snapshot(diff="does not exist index")



from typing import List

import pytest

# Constants for testing with extra plugin arguments
_NO_ARGS = []
_XDIST_ZERO = ["--numprocesses", "0"]
_XDIST_TWO = ["--numprocesses", "2"]


@pytest.fixture(
    params=[_NO_ARGS, _XDIST_ZERO, _XDIST_TWO],
    ids=["no_plugin", "xdist_zero", "xdist_two"],
)
def plugin_args(request: pytest.FixtureRequest) -> List[str]:
    """Fixture to test with various plugins"""
    return request.param


@pytest.fixture(
    params=[
        _NO_ARGS,
        _XDIST_ZERO,
        pytest.param(
            _XDIST_TWO,
            marks=pytest.mark.xfail(reason="Not currently compatible with xdist"),
        ),
    ],
    ids=["no_plugin", "xdist_zero", "xdist_two"],
)
def plugin_args_fails_xdist(request: pytest.FixtureRequest) -> List[str]:
    """Fixture to test with various plugins, but expected to fail xdist"""
    return request.param



import pytest


@pytest.fixture
def testcases():
    return {
        "used": (
            """
            def test_used(snapshot):
                assert snapshot == 'used'
            """
        ),
        "unused": (
            """
            def test_unused(snapshot):
                assert snapshot == 'unused'
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    pyfile_content = "\n\n".join(testcases.values())
    testdir.makepyfile(test_file=pyfile_content)
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"2 snapshots generated\.",))
    return testdir, testcases


def test_unused_snapshots_failure(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot unused\.",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 1


def test_unused_snapshots_warning(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest("-v", "--snapshot-warn-unused", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot unused\.",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 0


def test_unused_snapshots_deletion(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 unused snapshot deleted\.",
            r"Deleted test_unused \(__snapshots__[\\/]test_file\.ambr\)",
        )
    )
    assert result.ret == 0



import datetime
import uuid

import pytest

from syrupy.extensions.amber.serializer import (
    AmberDataSerializer,
    Repr,
)
from syrupy.matchers import (
    PathTypeError,
    compose_matchers,
    path_type,
    path_value,
)


def test_matcher_path_type_noop(snapshot):
    with pytest.raises(PathTypeError, match="argument cannot be empty"):
        path_type()


def test_matches_expected_type(snapshot):
    my_matcher = path_type(
        {"date_created": (datetime.datetime,), "nested.id": (int,)}, types=(uuid.UUID,)
    )
    actual = {
        "date_created": datetime.datetime.now(),
        "nested": {"id": 4},
        "nested_id": 5,
        "some_uuid": uuid.uuid4(),
    }
    assert actual == snapshot(matcher=my_matcher)


def test_raises_unexpected_type(snapshot):
    kwargs = {
        "mapping": {
            "date_created": (datetime.datetime,),
            "date_updated": (datetime.datetime,),
            "nested.id": (str,),
        },
        "types": (uuid.UUID, int),
    }
    actual = {
        "date_created": datetime.datetime.now(),
        "date_updated": datetime.date(2020, 6, 1),
        "nested": {"id": 4},
        "some_uuid": uuid.uuid4(),
    }
    assert actual == snapshot(matcher=path_type(**kwargs, strict=False))
    with pytest.raises(AssertionError, match="does not match any of the expected"):
        assert actual == snapshot(matcher=path_type(**kwargs))


def test_matches_non_deterministic_snapshots(snapshot):
    def matcher(data, path):
        if isinstance(data, uuid.UUID):
            return Repr("UUID(...)")
        if isinstance(data, datetime.datetime):
            return Repr("datetime.datetime(...)")
        if tuple(p for p, _ in path[-2:]) == ("c", 0):
            return "Your wish is my command"
        return data

    assert {
        "a": uuid.uuid4(),
        "b": {"b_1": "This is deterministic", "b_2": datetime.datetime.now()},
        "c": ["Replace this one", "Do not replace this one"],
    } == snapshot(matcher=matcher)
    assert {
        "a": uuid.UUID("06335e84-2872-4914-8c5d-3ed07d2a2f16"),
        "b": {
            "b_1": "This is deterministic",
            "b_2": datetime.datetime(year=2020, month=5, day=31),
        },
        "c": ["Replace this one", "Do not replace this one"],
    } == snapshot


def test_matches_regex_in_regex_mode(snapshot):
    my_matcher = path_type(
        {
            r"data\.list\..*\.date_created": (datetime.datetime,),
            r"any_number": (int,),
            "any_number.adjacent": (str,),
        },
        regex=True,
    )
    actual = {
        "data": {
            "list": [
                {"k": "1", "date_created": datetime.datetime.now()},
                {"k": "2", "date_created": datetime.datetime.now()},
            ],
        },
        "any_number": 3,
        "any_number_adjacent": "hi",
        "specific_number": 5,
    }
    assert actual == snapshot(matcher=my_matcher)


def test_regex_matcher_str_value(request, snapshot, tmp_path):
    def replacer(data, match):
        # check that the match is for the expected file path
        if match and request.node.name in match[0]:
            return match[0].replace(match[1], "<tmp-path>/")
        return Repr(AmberDataSerializer.object_type(data))

    my_matcher = path_value(
        {
            r"data\.list\..*\.id": "[a-z0-9]{8}-([a-z0-9]{4}-){3}[a-z0-9]{12}",
            "dir": rf"(.*){request.node.name}.*",
        },
        types=(str, uuid.UUID),
        replacer=replacer,
        regex=True,
    )
    actual = {
        "data": {
            "any_number": 3,
            "any_string": "hello",
            "list": [
                {"k": "1", "id": uuid.uuid4()},
                {"k": "2", "id": uuid.uuid4()},
            ],
        },
        "dir": str(tmp_path),
    }
    assert actual == snapshot(matcher=my_matcher)


def test_multiple_matchers(snapshot):
    data = {"number": 1, "datetime": datetime.datetime.now(), "float": 1.3}

    assert data == snapshot(
        matcher=compose_matchers(
            path_type(types=(list,), replacer=lambda *_: "DO_NOT_MATCH"),
            path_type(types=(int, float), replacer=lambda *_: "MATCHER_1"),
            path_type(types=(datetime.datetime,), replacer=lambda *_: "MATCHER_2"),
        ),
    )



import pytest


@pytest.fixture
def testcases():
    return {
        "a": (
            """
            def test_a(snapshot):
                assert snapshot == 'a'
            """
        ),
        "b": (
            """
            def test_b(snapshot):
                assert snapshot == 'b'
            """
        ),
        "a_suffix": (
            """
            def test_a_suffix(snapshot):
                assert snapshot == 'a_suffix'
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    pyfile_content = "\n\n".join(testcases.values())
    testdir.makepyfile(
        test_1=pyfile_content, test_2=pyfile_content, test_1_with_suffix=pyfile_content
    )
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"9 snapshots generated\.",))
    return testdir, testcases


def test_run_all(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(("9 snapshots passed",))
    assert result.ret == 0


def test_run_single_file(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest("-v", "test_1.py", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(("3 snapshots passed",))
    assert result.ret == 0


def test_run_single_test_case_in_file(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest("-v", "test_2.py::test_a", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(("1 snapshot passed",))
    assert result.ret == 0


def test_run_all_but_one(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "test_1.py",
        "test_2.py::test_a",
        *plugin_args_fails_xdist,
    )
    result.stdout.re_match_lines(("4 snapshots passed",))
    assert result.ret == 0


def test_run_both_files_by_node(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "test_1.py::test_a",
        "test_2.py::test_a",
        *plugin_args_fails_xdist,
    )
    result.stdout.re_match_lines(("2 snapshots passed",))
    assert result.ret == 0


def test_run_both_files_by_node_2(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "test_1.py::test_b",
        "test_2.py::test_a",
        *plugin_args_fails_xdist,
    )
    result.stdout.re_match_lines(("2 snapshots passed",))
    assert result.ret == 0



import pytest


@pytest.fixture
def testcases():
    return {
        "base": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case 2"
            """
        ),
        "modified": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case ??"
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    testdir.makepyfile(test_1=testcases["base"])
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
    )
    result.stdout.re_match_lines((r"2 snapshots generated\.",))
    return testdir, testcases


def test_run_all(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(("2 snapshots passed",))
    assert result.ret == 0


def test_failure(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_1=testcases["modified"])
    result = testdir.runpytest("-vv", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(("1 snapshot failed. 1 snapshot passed.",))
    assert result.ret == 1


def test_update(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_1=testcases["modified"])
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    assert "Can not relate snapshot name" not in str(result.stdout)
    result.stdout.re_match_lines(("1 snapshot passed. 1 snapshot updated.",))
    assert result.ret == 0



"""
Example: Custom Snapshot Directory

Here we extend the Amber extension to change the directory
in which snapshots are stored.

We explicitly name our new fixture "snapshot" to override the
default snapshot fixture. If this is placed in your project's
root conftest.py file, it is equivalent to globally overriding
the default snapshot directory.
"""

from pathlib import Path

import pytest

from syrupy.extensions.json import JSONSnapshotExtension
from syrupy.location import PyTestLocation


def create_versioned_fixture(version: int):
    class VersionedJSONExtension(JSONSnapshotExtension):
        @classmethod
        def dirname(cls, *, test_location: "PyTestLocation") -> str:
            return str(
                Path(test_location.filepath).parent.joinpath(
                    "__snapshots__", f"v{version}"
                )
            )

    return VersionedJSONExtension


VersionedJSONExtension1 = create_versioned_fixture(version=1)
VersionedJSONExtension2 = create_versioned_fixture(version=2)


@pytest.fixture
def snapshot_v1(snapshot):
    return snapshot.use_extension(VersionedJSONExtension1)


@pytest.fixture
def snapshot_v2(snapshot):
    return snapshot.use_extension(VersionedJSONExtension2)


def test_case_v1(snapshot_v1):
    assert "From v1 of an API" == snapshot_v1


def test_case_v2(snapshot_v2):
    assert "From v2 of an API" == snapshot_v2



import pytest


@pytest.fixture
def testcases():
    return {
        "used": (
            """
            def test_used(snapshot):
                assert snapshot == 'used'
            """
        ),
        "raise-skipped": (
            """
            import pytest
            def test_skipped(snapshot):
                pytest.skip("Skipping...")
                assert snapshot == 'unused'
            """
        ),
        "mark-skipped": (
            """
            import pytest
            @pytest.mark.skip
            def test_skipped(snapshot):
                assert snapshot == 'unused'
            """
        ),
        "not-skipped": (
            """
            def test_skipped(snapshot):
                assert snapshot == 'unused'
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    pyfile_content = "\n\n".join([testcases["used"], testcases["not-skipped"]])
    testdir.makepyfile(test_file=pyfile_content)
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines(r"2 snapshots generated\.")
    return testdir, testcases


def test_mark_skipped_snapshots(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    pyfile_content = "\n\n".join([testcases["used"], testcases["mark-skipped"]])
    testdir.makepyfile(test_file=pyfile_content)

    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(r"1 snapshot passed\.$")
    assert result.ret == 0


def test_raise_skipped_snapshots(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    pyfile_content = "\n\n".join([testcases["used"], testcases["raise-skipped"]])
    testdir.makepyfile(test_file=pyfile_content)

    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(r"1 snapshot passed\.$")
    assert result.ret == 0


def test_skipped_snapshots_update(run_testcases, plugin_args_fails_xdist):
    testdir, testcases = run_testcases
    pyfile_content = "\n\n".join([testcases["used"], testcases["raise-skipped"]])
    testdir.makepyfile(test_file=pyfile_content)

    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(r"1 snapshot passed\.$")
    assert result.ret == 0



import base64

import pytest

from syrupy.extensions.image import PNGImageSnapshotExtension

actual_png = base64.b64decode(
    b"iVBORw0KGgoAAAANSUhEUgAAADIAAAAyBAMAAADsEZWCAAAAG1BMVEXMzMy"
    b"Wlpaqqqq3t7exsbGcnJy+vr6jo6PFxcUFpPI/AAAACXBIWXMAAA7EAAAOxA"
    b"GVKw4bAAAAQUlEQVQ4jWNgGAWjgP6ASdncAEaiAhaGiACmFhCJLsMaIiDAE"
    b"QEi0WXYEiMCOCJAJIY9KuYGTC0gknpuHwXDGwAA5fsIZw0iYWYAAAAASUVO"
    b"RK5CYII="
)


@pytest.fixture
def snapshot_png(snapshot):
    return snapshot.use_extension(PNGImageSnapshotExtension)


def test_image(snapshot_png):
    assert actual_png == snapshot_png


def test_multiple_snapshot_extensions(snapshot):
    """
    Example of switching extension classes on the fly.
    These should be indexed in order of assertion.
    """
    assert actual_png == snapshot(extension_class=PNGImageSnapshotExtension)
    assert actual_png == snapshot()  # uses initial extension class
    assert snapshot._extension is not None
    assert actual_png == snapshot(extension_class=PNGImageSnapshotExtension)



import json
import os
from contextlib import contextmanager
from difflib import ndiff
from gettext import gettext
from importlib import import_module
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)

from .constants import (
    DIFF_LINE_COUNT_LIMIT,
    DIFF_LINE_WIDTH_LIMIT,
    SYMBOL_ELLIPSIS,
)
from .exceptions import FailedToLoadModuleMember


def is_xdist_worker() -> bool:
    worker_name = os.getenv("PYTEST_XDIST_WORKER")
    return bool(worker_name and worker_name != "master")


def is_xdist_controller() -> bool:
    worker_count = os.getenv("PYTEST_XDIST_WORKER_COUNT")
    return bool(worker_count and int(worker_count) > 0 and not is_xdist_worker())


def walk_snapshot_dir(
    root: Union[str, Path], *, ignore_extensions: Optional[List[str]] = None
) -> Iterator[str]:
    ignore_exts: set[str] = set(ignore_extensions or [])

    for filepath in Path(root).rglob("*"):
        if not filepath.name.startswith(".") and filepath.is_file():
            if filepath.suffixes and filepath.suffixes[-1][1:] in ignore_exts:
                continue
            yield str(filepath)


def import_module_member(path: str) -> Any:
    sep = "."
    [*module_parts, module_member_name] = path.split(sep)
    module_name = sep.join(module_parts)

    if not module_name:
        raise FailedToLoadModuleMember(
            gettext("Cannot load member '{}' without module path").format(
                module_member_name,
            )
        )
    try:
        module = import_module(module_name)
    except ModuleNotFoundError as e:
        raise FailedToLoadModuleMember(
            gettext("Module '{}' does not exist.").format(module_name)
        ) from e

    try:
        return getattr(module, module_member_name)
    except AttributeError as e:
        raise FailedToLoadModuleMember(
            gettext("Member '{}' not found in module '{}'.").format(
                module_member_name,
                module_name,
            )
        ) from e


def get_env_value(env_var_name: str) -> object:
    try:
        return json.loads(os.environ[env_var_name])
    except (KeyError, TypeError, json.decoder.JSONDecodeError):
        return os.environ.get(env_var_name)


@contextmanager
def env_context(**kwargs: str) -> Iterator[None]:
    prev_env = {**os.environ}
    try:
        yield os.environ.update(kwargs)
    finally:
        os.environ.clear()
        os.environ.update(prev_env)


def set_attrs(obj: Any, attrs: Dict[str, Any]) -> Any:
    for k in attrs:
        setattr(obj, k, attrs[k])


@contextmanager
def obj_attrs(obj: Any, attrs: Dict[str, Any]) -> Iterator[None]:
    prev_attrs = {k: getattr(obj, k, None) for k in attrs}
    try:
        yield set_attrs(obj, attrs)
    finally:
        set_attrs(obj, prev_attrs)


def qdiff(
    lines_a: "Sequence[str]",
    lines_b: "Sequence[str]",
    *,
    line_diff_limit: int = DIFF_LINE_COUNT_LIMIT,
    line_size_limit: int = DIFF_LINE_WIDTH_LIMIT,
) -> "Iterator[str]":
    """
    Wrapper around difflib ndiff to bail early
    https://github.com/python/cpython/issues/65452
    """
    first_diff_line_idx = 0
    first_diff_char_idx = 0

    for i in range(max(len(lines_a), len(lines_b))):
        line_a = "".join(lines_a[i : i + 1])  # noqa E203
        line_b = "".join(lines_b[i : i + 1])  # noqa E203
        if line_a != line_b:
            first_diff_line_idx = i
            for j in range(max(len(line_a), len(line_b))):
                char_a = line_a[j : j + 1]  # noqa E203
                char_b = line_b[j : j + 1]  # noqa E203
                if char_a != char_b:
                    first_diff_char_idx = j
                    break
            break

    def adjust_lines(lines: "Sequence[str]") -> "Sequence[str]":
        line_idx_from = max(first_diff_line_idx - line_diff_limit, 0)
        line_idx_to = first_diff_line_idx + line_diff_limit

        symbol_hidden_line = SYMBOL_ELLIPSIS + SYMBOL_ELLIPSIS
        return (
            # include an indicator in the diff if this was not the first line
            ([symbol_hidden_line] if line_idx_from > 0 else [])
            # show included lines with the ends truncated off
            + [
                adj_line
                for n, line in enumerate(lines[line_idx_from:line_idx_to])
                # adjust the first line shown to be from the first different spotted
                for line_start, line_end in [
                    (
                        (
                            max(first_diff_char_idx - line_size_limit, 0)
                            if n == line_idx_from
                            else 0
                        ),
                        (
                            first_diff_char_idx + line_size_limit
                            if n == line_idx_from
                            else line_size_limit
                        ),
                    ),
                ]
                for adj_line in [
                    (SYMBOL_ELLIPSIS if line_start > 0 else "")
                    + line[line_start:line_end]
                    + (SYMBOL_ELLIPSIS if line_end < len(line) else "")
                ]
            ]
            # include an indicator in the diff if this was not the last line
            + ([symbol_hidden_line] if line_idx_to < len(lines) else [])
        )

    return ndiff(adjust_lines(lines_a), adjust_lines(lines_b))



from invoke import (
    exceptions,
    task,
)

from .utils import ctx_run

lint_commands = {
    "mypy": lambda _: "python -m mypy --strict src benchmarks",
    "ruff": lambda fix: f"python -m ruff check {'--fix' if fix else ''} .",
    "ruff-format": lambda fix: f"python -m ruff format {'' if fix else '--diff'} .",
}


def run_lint(ctx, section, fix):
    print(f"\033[1m[{section}]\033[0m")
    ctx_run(ctx, lint_commands[section](fix))


@task(default=True)
def all(ctx, fix=False):  # noqa: A001
    """
    Check and fix syntax using various linters
    """
    last_error = None
    for section in lint_commands:
        try:
            run_lint(ctx, section, fix)
            print()
        except exceptions.Failure as ex:
            last_error = ex
    if last_error:
        raise last_error


def add_individual_task(name: str):
    @task(name=name)
    def lint(ctx, fix=False):
        run_lint(ctx, name, fix)

    globals()[name] = lint


for section in lint_commands:
    add_individual_task(section)



from invoke import task


@task()
def benchmark(ctx, report=False):
    """
    Run and generate benchmarks for current code
    """

    ctx.run("pytest benchmarks --benchmark-json=benchmarks.json")



# noqa: A005
from typing import (
    Any,
    Callable,
    Hashable,
    Optional,
    Tuple,
    Type,
    Union,
)

SnapshotIndex = Union[int, str]
SerializableData = Any
SerializedData = Union[str, bytes]
PropertyName = Hashable
PropertyValueType = Type[SerializableData]
PropertyPathEntry = Tuple[PropertyName, PropertyValueType]
PropertyPath = Tuple[PropertyPathEntry, ...]
try:
    from mypy_extensions import NamedArg

    PropertyMatcher = Callable[
        [
            NamedArg(SerializableData, "data"),  # noqa: F821
            NamedArg(PropertyPath, "path"),  # noqa: F821
        ],
        Optional[SerializableData],
    ]
    PropertyFilter = Callable[
        [
            NamedArg(PropertyName, "prop"),  # noqa: F821
            NamedArg(PropertyPath, "path"),  # noqa: F821
        ],
        bool,
    ]

except ImportError:
    globals()["PropertyMatcher"] = Callable[..., Optional[SerializableData]]
    globals()["PropertyFilter"] = Callable[..., bool]



import pytest


@pytest.fixture(params=["__snapshots__", "snapshots"])
def testcases(testdir, tmp_path, request):
    dirname = tmp_path.joinpath(request.param)
    testdir.makeconftest(
        f"""
        import pytest

        from syrupy.extensions.amber import AmberSnapshotExtension

        class CustomSnapshotExtension(AmberSnapshotExtension):
            @classmethod
            def dirname(cls, *, test_location):
                return {str(dirname)!r}

        @pytest.fixture
        def snapshot(snapshot):
            return snapshot.use_extension(CustomSnapshotExtension)
        """
    )
    return {
        "zero": (
            """
            def test_do_it(snapshot):
                pass
            """
        ),
        "one": (
            """
            def test_do_it(snapshot):
                assert snapshot == 'passed1'
            """
        ),
        "two": (
            """
            def test_do_it(snapshot):
                assert snapshot == 'passed1'
                assert snapshot == 'passed2'
            """
        ),
    }


@pytest.fixture
def generate_snapshots(testdir, testcases):
    testdir.makepyfile(test_file=testcases["two"])
    result = testdir.runpytest("-v", "--snapshot-update")
    return result, testdir, testcases


def test_generated_snapshots(generate_snapshots):
    result = generate_snapshots[0]
    result.stdout.re_match_lines((r"2 snapshots generated\.",))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0


def test_unmatched_snapshots(generate_snapshots, plugin_args_fails_xdist):
    _, testdir, testcases = generate_snapshots
    testdir.makepyfile(test_file=testcases["one"])
    result = testdir.runpytest("-v", *plugin_args_fails_xdist)
    result.stdout.re_match_lines((r"1 snapshot passed. 1 snapshot unused\.",))
    assert result.ret == 1


def test_updated_snapshots_partial_delete(generate_snapshots, plugin_args_fails_xdist):
    _, testdir, testcases = generate_snapshots
    testdir.makepyfile(test_file=testcases["one"])
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(r"1 snapshot passed. 1 unused snapshot deleted\.")
    assert result.ret == 0


def test_updated_snapshots_full_delete(generate_snapshots, plugin_args_fails_xdist):
    _, testdir, testcases = generate_snapshots
    testdir.makepyfile(test_file=testcases["zero"])
    result = testdir.runpytest("-v", "--snapshot-update", *plugin_args_fails_xdist)
    result.stdout.re_match_lines(r"2 unused snapshots deleted\.")
    assert result.ret == 0



import pytest

from syrupy.constants import DIFF_LINE_COUNT_LIMIT
from syrupy.extensions.base import SnapshotReporter


class SnapshotReporterNoContext(SnapshotReporter):
    @property
    def _context_line_count(self) -> int:
        return 0


@pytest.mark.parametrize("Reporter", [SnapshotReporter, SnapshotReporterNoContext])
class TestSnapshotReporter:
    @pytest.mark.parametrize(
        "a, b",
        [
            (
                "line 0\nline 1\nline 02\nline 3\nline 4\r\nline 5\nline 6\nline 7",
                "line 0\nline 1\nline 2\r\nline 3\nline 04\nline 5\nline 6\nline 7",
            ),
            (
                "line 0\nline 1\nline 2\nline 3\t\nline 4\nline 5\nline 6\nline 7",
                "line 0\nline 1\nline 2\nline 3  \nline 4\nline 5\nline 6\nline 7",
            ),
            (
                "line 0\nline 1\nline 2\nline 3\r\nline 4\nline 5\nline 6\nline 7",
                "line 0\rline 1\nline 2\r\nline 3\nline 4\nline 5\nline 6\nline 7",
            ),
        ],
        ids=lambda _: "",
    )
    def test_diff_lines(self, a, b, Reporter, snapshot, osenv):
        with osenv(NO_COLOR="true"):
            assert "\n".join(Reporter().diff_lines(a, b)) == snapshot

    def test_diff_large(self, Reporter, snapshot, osenv):
        n_count = 3000 + DIFF_LINE_COUNT_LIMIT * 2
        obj_a = [str(x) + ("a" * 20) for x in range(n_count)]
        obj_b = [
            line_a + ("b" if n > DIFF_LINE_COUNT_LIMIT else "")
            for n, line_a in enumerate(obj_a)
        ]
        str_a = "\n".join(obj_a)
        str_b = "\n".join(obj_b)
        with osenv(NO_COLOR="true"):
            assert "\n".join(Reporter().diff_lines(str_a, str_b)) == snapshot

    def test_diff_large_lines(self, Reporter, snapshot, osenv):
        n_count = 1000000
        obj_a = [str(x) + ("a" * n_count) for x in range(20)]
        obj_b = [line_a[: n_count // 2] + "b" * n_count for line_a in obj_a]
        str_a = "\n".join(obj_a)
        str_b = "\n".join(obj_b)
        with osenv(NO_COLOR="true"):
            assert "\n".join(Reporter().diff_lines(str_a, str_b)) == snapshot



import datetime

import pytest

from syrupy.extensions.json import JSONSnapshotExtension
from syrupy.filters import (
    paths,
    props,
)


@pytest.fixture
def snapshot_json(snapshot):
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.mark.parametrize(
    "content",
    [
        {},
        ["an array"],
        {
            "str": "foo",
            "int": -1,
            "float": 4.2,
            "array": [1, 2, 3],
            "null": None,
            "datetime": datetime.datetime(2021, 1, 31, 23, 59),
        },
    ],
)
def test_serializer(snapshot_json, content):
    assert snapshot_json == content


def test_exclude_simple(snapshot_json):
    content = {
        "id": 123456789,
        "foo": "__SHOULD_BE_REMOVED_FROM_JSON__",
        "I'm": "still alive",
        "nested": {
            "foo": "is still alive",
        },
    }
    assert snapshot_json(exclude=props("id", "foo")) == content
    assert snapshot_json(exclude=paths("id", "foo")) == content


def test_include_simple(snapshot_json):
    content = {
        "id": 123456789,
        "foo": "__SHOULD_BE_REMOVED_FROM_JSON__",
        "I'm": "still alive",
        "nested": {
            "foo": "is still alive",
        },
    }
    assert snapshot_json(include=props("id", "foo")) == content
    assert snapshot_json(include=paths("id", "foo")) == content


def test_exclude_nested(snapshot_json):
    content = {
        "a": "b",
        "foo": {
            "bar": "__SHOULD_BE_REMOVED_FROM_JSON__",
        },
        "x": {
            "y": {
                "z": "__SHOULD_BE_REMOVED_FROM_JSON__",
                "zz": "I'm still there",
            }
        },
    }
    assert snapshot_json(exclude=paths("foo.bar", "x.y.z")) == content


def test_exclude_in_json_with_empty_values(snapshot_json):
    content = {
        "foo": "bar",
        "none": None,
        "empty_dict": {},
        "empty_list": [],
    }
    assert snapshot_json(exclude=props("foo")) == content



def doctest_fn():
    """a doctest in a function docstring
    >>> doctest_fn() == getfixture('snapshot')
    True
    """
    return "doc test fn return value"


class DocTestClass:
    """
    >>> DocTestClass() == getfixture('snapshot')
    True

    a doctest in a class docstring
    >>> DocTestClass() == getfixture('snapshot')
    True
    """

    obj_attr = "test class attr"

    def doctest_method(self):
        """a doctest in a method docstring
        >>> DocTestClass().doctest_method() == getfixture('snapshot')
        True
        """
        return "doc test method return value"

    class NestedDocTestClass:
        """a doctest in a nested class docstring
        >>> DocTestClass.NestedDocTestClass() == getfixture('snapshot')
        True
        """

        nested_obj_attr = "nested doc test class attr"

        def doctest_method(self):
            """a doctest in a nested method docstring
            >>> nested_obj = DocTestClass.NestedDocTestClass()
            >>> nested_obj.doctest_method() == getfixture('snapshot')
            True
            """
            return "nested doc test method return value"



"""Example: Custom snapshot defaults

Here we modify snapshot defaults globally, instead of per assert.
This gives the ability to modify the snapshot functionality to your hearts content
and then simply re-use them, without having to pass those defaults to every assert.
Especially useful if there's a lot of tests that need to modify the default behaviour.
"""

import pytest

from syrupy.extensions.json import JSONSnapshotExtension
from syrupy.filters import paths
from syrupy.matchers import path_type


@pytest.fixture
def snapshot_matcher(snapshot):
    return snapshot.with_defaults(matcher=path_type(mapping={"my-field": (str, int)}))


@pytest.fixture
def snapshot_exclude(snapshot_matcher):
    return snapshot_matcher.with_defaults(exclude=paths("excluded"))


@pytest.fixture
def snapshot_json(snapshot_exclude):
    return snapshot_exclude.with_defaults(extension_class=JSONSnapshotExtension)


def test_asserting_multiple_times(snapshot_matcher):
    assert {"my-field": "my-string", "excluded": "value"} == snapshot_matcher
    assert {"my-field": 12345, "excluded": "value"} == snapshot_matcher


def test_asserting_multiple_times_chained(snapshot_exclude):
    assert {"my-field": "my-string", "excluded": "value"} == snapshot_exclude
    assert {"my-field": 12345, "excluded": "value"} == snapshot_exclude


def test_asserting_multiple_times_chained_json(snapshot_json):
    assert {"my-field": "my-string", "excluded": "value"} == snapshot_json
    assert {"my-field": 12345, "excluded": "value"} == snapshot_json



"""
Example: Custom Image Extension

The Single File Snapshot Extension writes each snapshot
to its own file. For images, this produces a cleaner diff
on websites like GitHub.
"""

import base64

import pytest

from syrupy.extensions.single_file import SingleFileSnapshotExtension


class JPEGImageExtension(SingleFileSnapshotExtension):
    _file_extension = "jpg"


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(JPEGImageExtension)


def test_jpeg_image(snapshot):
    reddish_square = base64.b64decode(
        b"/9j/4AAQSkZJRgABAQEASABIAAD/2wBDAP//////////////////////////"
        b"////////////////////////////////////////////////////////////"
        b"2wBDAf//////////////////////////////////////////////////////"
        b"////////////////////////////////wAARCAAEAAQDASIAAhEBAxEB/8QA"
        b"HwAAAQUBAQEBAQEAAAAAAAAAAAECAwQFBgcICQoL/8QAtRAAAgEDAwIEAwUF"
        b"BAQAAAF9AQIDAAQRBRIhMUEGE1FhByJxFDKBkaEII0KxwRVS0fAkM2JyggkK"
        b"FhcYGRolJicoKSo0NTY3ODk6Q0RFRkdISUpTVFVWV1hZWmNkZWZnaGlqc3R1"
        b"dnd4eXqDhIWGh4iJipKTlJWWl5iZmqKjpKWmp6ipqrKztLW2t7i5usLDxMXG"
        b"x8jJytLT1NXW19jZ2uHi4+Tl5ufo6erx8vP09fb3+Pn6/8QAHwEAAwEBAQEB"
        b"AQEBAQAAAAAAAAECAwQFBgcICQoL/8QAtREAAgECBAQDBAcFBAQAAQJ3AAEC"
        b"AxEEBSExBhJBUQdhcRMiMoEIFEKRobHBCSMzUvAVYnLRChYkNOEl8RcYGRom"
        b"JygpKjU2Nzg5OkNERUZHSElKU1RVVldYWVpjZGVmZ2hpanN0dXZ3eHl6goOE"
        b"hYaHiImKkpOUlZaXmJmaoqOkpaanqKmqsrO0tba3uLm6wsPExcbHyMnK0tPU"
        b"1dbX2Nna4uPk5ebn6Onq8vP09fb3+Pn6/9oADAMBAAIRAxEAPwBlFFFAz//Z"
    )
    assert reddish_square == snapshot



import re
from gettext import gettext
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
)

from syrupy.extensions.amber.serializer import (
    AmberDataSerializer,
    Repr,
)

if TYPE_CHECKING:
    from syrupy.types import (
        PropertyMatcher,
        PropertyPath,
        PropertyValueType,
        SerializableData,
    )

    try:
        MatchResult = Optional[re.Match[str]]
    except TypeError:
        globals()["MatchResult"] = Optional[re.Match]
    Replacer = Callable[[SerializableData, MatchResult], SerializableData]


class PathTypeError(TypeError):
    pass


class StrictPathTypeError(PathTypeError):
    pass


def path_type(
    mapping: Optional[Dict[str, Tuple["PropertyValueType", ...]]] = None,
    *,
    types: Tuple["PropertyValueType", ...] = (),
    strict: bool = True,
    regex: bool = False,
    replacer: "Replacer" = lambda data, _: Repr(AmberDataSerializer.object_type(data)),
) -> "PropertyMatcher":
    """
    Factory to create a matcher using path and type mapping.
    Usecase:
    Replacing all values of certain types at specified paths.
    with their class name instead. Allows for deterministic
    snapshots on non-deterministic class e.g. datetime, random etc.
    """
    if not mapping and not types:
        raise PathTypeError(gettext("Both mapping and types argument cannot be empty"))

    def path_type_matcher(
        *, data: "SerializableData", path: "PropertyPath"
    ) -> Optional["SerializableData"]:
        path_str = ".".join(str(p) for p, _ in path)
        if mapping:
            for pattern in mapping:
                matches = _path_match(path_str, pattern, regex)
                if matches:
                    for type_to_match in mapping[pattern]:
                        if isinstance(data, type_to_match):
                            return replacer(data, matches)
                    if strict:
                        raise StrictPathTypeError(
                            gettext(
                                "{} at '{}' of type {} does not "
                                "match any of the expected types: {}"
                            ).format(data, path_str, data.__class__, mapping[pattern])
                        )
        for type_to_match in types:
            if isinstance(data, type_to_match):
                return replacer(data, None)
        return data

    return path_type_matcher


def path_value(
    mapping: Dict[str, str],
    *,
    types: Tuple["PropertyValueType", ...],
    replacer: "Replacer",
    **kwargs: Any,
) -> "PropertyMatcher":
    """
    Matches the path regex against the string repr of values for the types specified
    """

    kwargs["mapping"] = {path_pattern: types for path_pattern in mapping}
    kwargs["replacer"] = lambda data, path_matches: (
        replacer(
            data,
            _path_match(
                str(data), mapping[path_matches.re.pattern], kwargs.get("regex", False)
            ),
        )
        if path_matches.re.pattern in mapping
        else data
    )
    return path_type(**kwargs)


def _path_match(path: str, pattern: str, is_regex: bool) -> "MatchResult":
    """Match path against regular string or regex pattern"""
    if not is_regex:
        pattern = re.escape(pattern)
    return re.fullmatch(pattern, path)


def compose_matchers(*matchers: "PropertyMatcher") -> "PropertyMatcher":
    """
    Composes 1 or more matchers into a single matcher.
    """

    def _matcher(
        *, data: "SerializableData", path: "PropertyPath"
    ) -> Optional["SerializableData"]:
        for matcher in matchers:
            try:
                data = matcher(data=data, path=path)
            except StrictPathTypeError:
                # ignore strict mode when composing matchers
                pass

        return data

    return _matcher



# type: ignore
import shutil
from pathlib import Path


def test_1000x_reads(testdir, benchmark):
    test_contents = """
        import pytest

        @pytest.mark.parametrize("x", range(1000))
        def test_performance(x, snapshot):
            assert x == snapshot
        """
    testdir.makepyfile(test=test_contents)

    # This test benchmarks reads not writes, so we'll pre-write the snapshots.
    testdir.runpytest("test.py", "--snapshot-update")

    # Run benchmark
    benchmark(lambda: testdir.runpytest("test.py"))


def test_1000x_writes(testdir, benchmark):
    test_contents = """
        import pytest

        @pytest.mark.parametrize("x", range(1000))
        def test_performance(x, snapshot):
            assert x == snapshot
        """

    def fn():
        test_path = testdir.makepyfile(test=test_contents)
        testdir.runpytest("test.py", "--snapshot-update")
        shutil.rmtree(Path(test_path).parent / "__snapshots__", ignore_errors=True)

    # Run benchmark
    benchmark(fn)
